# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import array
import base64
import datetime
import os

import future.utils

try:
    from urllib.parse import quote, urlencode
except BaseException:
    from urllib import urlencode, quote

import ujson

import v3io.common.helpers
import v3io.dataplane.kv_array
import v3io.dataplane.kv_timestamp

#
# Request
#


class Request(object):
    def __init__(self, container, access_key, raise_for_status, encoder, encoder_args, output=None):
        self.container = container
        self.access_key = access_key
        self.raise_for_status = raise_for_status
        self.encoder = encoder
        self.encoder_args = encoder_args
        self.output = output

        # get request params with the encoder
        self.method, self.path, self.query, self.headers, self.body = encoder(container, access_key, encoder_args)

        # used by the transport
        self.transport = lambda: None

    def encode_path(self):
        if self.query is None:
            return quote(self.path)

        return quote(self.path) + "?" + urlencode(self.query, quote_via=quote)


#
# Encoders
#


#
# Container
#


def encode_get_containers(container_name, access_key, kwargs):
    return _encode("GET", "/", access_key, None, None, {}, None)


def encode_get_container_contents(container_name, access_key, kwargs):
    query = {"prefix": kwargs["path"]}

    if kwargs["get_all_attributes"]:
        query["prefix-info"] = 1

    if kwargs["directories_only"]:
        query["prefix-only"] = 1

    if kwargs["limit"] is not None:
        query["max-keys"] = kwargs["limit"]

    if kwargs["marker"]:
        query["marker"] = kwargs["marker"]

    return _encode("GET", "/" + container_name, access_key, None, query, {}, None)


#
# Object
#


def encode_head_object(container_name, access_key, kwargs):
    return _encode("HEAD", container_name, access_key, kwargs["path"], None, None, None)


def encode_get_object(container_name, access_key, kwargs):
    headers = None
    offset = kwargs.get("offset")
    num_bytes = kwargs.get("num_bytes")

    # if the offset/num_bytes param is specified, add a range header
    if offset or num_bytes:
        offset = offset or 0
        range_value = "bytes=" + str(offset)

        if num_bytes:
            range_value += "-" + str(offset + num_bytes - 1)

        headers = {"Range": range_value}

    return _encode("GET", container_name, access_key, kwargs["path"], None, headers, None)


def encode_put_object(container_name, access_key, kwargs):
    headers = None

    # if the append flag is passed, add a range header
    if kwargs["append"]:
        headers = {"Range": "-1"}

    return _encode("PUT", container_name, access_key, kwargs["path"], None, headers, kwargs["body"])


def encode_delete_object(container_name, access_key, kwargs):
    return _encode("DELETE", container_name, access_key, kwargs["path"], None, None, None)


#
# KV
#


def encode_put_item(container_name, access_key, kwargs):
    # add 'Item' to body
    body = {"Item": _dict_to_typed_attributes(kwargs["attributes"])}

    if kwargs["condition"] is not None:
        body["ConditionExpression"] = kwargs["condition"]

    return _encode(
        "PUT",
        container_name,
        access_key,
        kwargs.get("path") or os.path.join(kwargs["table_path"], kwargs["key"]),
        None,
        {"X-v3io-function": "PutItem"},
        body,
    )


def encode_update_item(container_name, access_key, kwargs):
    body = {"UpdateMode": kwargs.get("update_mode") or "CreateOrReplaceAttributes"}

    if kwargs["condition"] is not None:
        body["ConditionExpression"] = kwargs["condition"]

    if not kwargs["expression"] and not kwargs["attributes"]:
        raise RuntimeError("One of expression or attributes must be populated for update item")

    if kwargs["expression"]:
        http_method = "POST"
        function_name = "UpdateItem"
        body["UpdateExpression"] = kwargs["expression"]

    if kwargs["alternate_expression"]:
        http_method = "POST"
        function_name = "UpdateItem"
        body["AlternateUpdateExpression"] = kwargs["alternate_expression"]

    elif kwargs["attributes"]:
        http_method = "PUT"
        function_name = "PutItem"
        body["Item"] = _dict_to_typed_attributes(kwargs["attributes"])

    return _encode(
        http_method,
        container_name,
        access_key,
        kwargs.get("path") or os.path.join(kwargs["table_path"], kwargs["key"]),
        None,
        {"X-v3io-function": function_name},
        body,
    )


def encode_get_item(container_name, access_key, kwargs):
    body = {"AttributesToGet": ",".join(kwargs["attribute_names"])}

    return _encode(
        "PUT",
        container_name,
        access_key,
        kwargs.get("path") or os.path.join(kwargs["table_path"], kwargs["key"]),
        None,
        {"X-v3io-function": "GetItem"},
        body,
    )


def encode_get_items(container_name, access_key, kwargs):
    body = {
        "AttributesToGet": ",".join(kwargs["attribute_names"]),
    }

    if kwargs.get("table_name"):
        body["TableName"] = kwargs["table_name"]

    if kwargs["filter_expression"]:
        body["FilterExpression"] = kwargs["filter_expression"]

    if kwargs["marker"]:
        body["Marker"] = kwargs["marker"]

    if kwargs["sharding_key"]:
        body["ShardingKey"] = kwargs["sharding_key"]

    if kwargs["limit"] is not None:
        body["Limit"] = kwargs["limit"]

    if kwargs["segment"] is not None:
        body["Segment"] = kwargs["segment"]

    if kwargs["total_segments"] is not None:
        body["TotalSegment"] = kwargs["total_segments"]

    if kwargs["sort_key_range_start"]:
        body["SortKeyRangeStart"] = kwargs["sort_key_range_start"]

    if kwargs["sort_key_range_end"]:
        body["SortKeyRangeEnd"] = kwargs["sort_key_range_end"]

    return _encode(
        "PUT",
        container_name,
        access_key,
        kwargs.get("path") or kwargs["table_path"],
        None,
        {"X-v3io-function": "GetItems"},
        body,
    )


#
# Stream
#


def encode_create_stream(container_name, access_key, kwargs):
    body = {"ShardCount": kwargs["shard_count"], "RetentionPeriodHours": kwargs.get("retention_period_hours") or 24}

    return _encode(
        "POST",
        container_name,
        access_key,
        kwargs.get("path") or kwargs["stream_path"],
        None,
        {"X-v3io-function": "CreateStream"},
        body,
    )


def encode_update_stream(container_name, access_key, kwargs):
    body = {
        "ShardCount": kwargs["shard_count"],
    }

    return _encode(
        "POST",
        container_name,
        access_key,
        kwargs.get("path") or kwargs["stream_path"],
        None,
        {"X-v3io-function": "UpdateStream"},
        body,
    )


def encode_describe_stream(container_name, access_key, kwargs):
    return _encode(
        "PUT",
        container_name,
        access_key,
        kwargs.get("path") or kwargs["stream_path"],
        None,
        {"X-v3io-function": "DescribeStream"},
        None,
    )


def encode_seek_shard(container_name, access_key, kwargs):
    body = {
        "Type": kwargs["seek_type"],
    }

    if kwargs["seek_type"] == "SEQUENCE":
        body["StartingSequenceNumber"] = kwargs["starting_sequence_number"]
    elif kwargs["seek_type"] == "TIME":
        body["TimestampSec"] = kwargs["timestamp_sec"]
        body["TimestampNSec"] = kwargs["timestamp_nsec"]
    elif kwargs["seek_type"] not in ["EARLIEST", "LATEST"]:
        raise ValueError(
            "Unsupported seek_type ({0}) for seek_shard. Must be one of SEQUENCE, TIME, EARLIEST, LATEST".format(
                kwargs["seek_type"]
            )
        )

    return _encode(
        "PUT",
        container_name,
        access_key,
        kwargs.get("path") or kwargs["stream_path"],
        None,
        {"X-v3io-function": "SeekShard"},
        body,
    )


def encode_put_records(container_name, access_key, kwargs):
    records = []

    for record in kwargs["records"]:
        record_body = {
            "Data": _to_base64(record["data"]),
        }

        try:
            record_body["ClientInfo"] = _to_base64(record["client_info"])
        except KeyError:
            pass

        try:
            record_body["ShardId"] = record["shard_id"]
        except KeyError:
            pass

        try:
            record_body["PartitionKey"] = record["partition_key"]
        except KeyError:
            pass

        records.append(record_body)

    body = {"Records": records}

    return _encode(
        "POST",
        container_name,
        access_key,
        kwargs.get("path") or kwargs["stream_path"],
        None,
        {"X-v3io-function": "PutRecords"},
        body,
    )


def encode_get_records(container_name, access_key, kwargs):
    body = {
        "Location": kwargs["location"],
    }

    limit = kwargs.get("limit")
    if limit:
        body["Limit"] = limit

    return _encode(
        "PUT",
        container_name,
        access_key,
        kwargs.get("path") or kwargs["stream_path"],
        None,
        {"X-v3io-function": "GetRecords"},
        body,
    )


#
# Helpers
#


def _encode(method, container_name, access_key, path, query, headers, body):
    if path is not None:
        path = v3io.common.helpers.url_join(container_name, path)
    else:
        path = container_name

    headers, body = _resolve_body_and_headers(access_key, headers, body)

    return method, path, query, headers, body


def _typed_attributes_to_dict(self):
    pass


def _ensure_trailing_slash(path):
    if not path.endswith("/"):
        return path + "/"

    return path


def _to_base64(input):
    if isinstance(input, str):
        input = input.encode("utf-8")

    return base64.b64encode(input)


def _dict_to_typed_attributes(d):
    typed_attributes = {}

    for key, value in future.utils.viewitems(d):
        attribute_type = type(value)
        type_value = None

        if isinstance(value, future.utils.text_type):
            type_key = "S"
            type_value = value
        elif isinstance(value, future.utils.string_types):
            type_key = "S"
            type_value = str(value)
        elif attribute_type in [int, float]:
            type_key = "N"
            type_value = str(value)
        elif attribute_type in [bytes, bytearray]:
            type_key = "B"
            type_value = base64.b64encode(value)
        elif isinstance(value, bool):
            type_key = "BOOL"
            type_value = value
        elif isinstance(value, list):
            type_key = "B"
            type_value = v3io.dataplane.kv_array.encode_list(value)
        elif isinstance(value, array.array):
            type_key = "B"
            type_value = v3io.dataplane.kv_array.encode_array(value, value.typecode)
        elif isinstance(value, datetime.datetime):
            type_key = "TS"
            type_value = v3io.dataplane.kv_timestamp.encode(value)
        else:
            raise AttributeError("Attribute {0} has unsupported type {1}".format(key, attribute_type))

        typed_attributes[key] = {type_key: type_value}

    return typed_attributes


def _resolve_body_and_headers(access_key, headers, body):
    if access_key:
        headers = headers or {}
        headers["X-v3io-session-key"] = access_key

    if not isinstance(body, dict):
        return headers, body

    body = ujson.dumps(body, reject_bytes=False)
    headers["Content-Type"] = "application/json"

    return headers, body



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import http.client
import queue
import socket
import ssl
import sys

import v3io.dataplane.request
import v3io.dataplane.response

from . import abstract


class Transport(abstract.Transport):
    def __init__(self, logger, endpoint=None, max_connections=None, timeout=None, verbosity=None):
        super(Transport, self).__init__(logger, endpoint, max_connections, timeout, verbosity)

        self._free_connections = queue.Queue()

        # based on scheme, create a host and context for _create_connection
        self._host, self._ssl_context = self._parse_endpoint(self._endpoint)

        # create the pool connection
        self._create_connections(self.max_connections, self._host, self._ssl_context)

        # python 2 and 3 have different exceptions
        if sys.version_info[0] >= 3:
            self._wait_response_exceptions = (
                http.client.RemoteDisconnected,
                ConnectionResetError,
                ConnectionRefusedError,
                http.client.ResponseNotReady,
            )
            self._send_request_exceptions = (
                BrokenPipeError,
                http.client.CannotSendRequest,
                http.client.RemoteDisconnected,
            )
            self._get_status_and_headers = self._get_status_and_headers_py3
        else:
            self._wait_response_exceptions = (http.client.BadStatusLine, socket.error)
            self._send_request_exceptions = (http.client.CannotSendRequest, http.client.BadStatusLine)
            self._get_status_and_headers = self._get_status_and_headers_py2

    def close(self):
        # Ignore redundant calls to close
        if not self._free_connections:
            return

        connections = []
        while not self._free_connections.empty():
            conn = self._free_connections.get()
            connections.append(conn)
        # In case anyone tries to reuse this object, we want them to get an error and not hang
        self._free_connections = None
        self._logger.debug(f"Closing all {len(connections)} v3io transport connections")
        for conn in connections:
            conn.close()

    def requires_access_key(self):
        return True

    def send_request(self, request):
        if not self._free_connections:
            raise RuntimeError("Cannot send request on a closed client")

        # TODO: consider getting param of whether we should block or
        #       not (wait for connection to be free or raise exception)
        connection = self._free_connections.get(block=True, timeout=None)

        try:
            return self._send_request_on_connection(request, connection)
        except BaseException as e:
            connection.close()
            connection = self._create_connection(self._host, self._ssl_context)
            self._free_connections.put(connection, block=True)
            raise e

    def wait_response(self, request, raise_for_status=None, num_retries=1):
        connection = request.transport.connection_used
        is_retry = False

        while True:
            response_body = None
            status_code = None
            headers = None
            try:
                if is_retry:
                    request = self._send_request_on_connection(request, connection)
                    connection = request.transport.connection_used

                response = connection.getresponse()
                response_body = response.read()

                status_code, headers = self._get_status_and_headers(response)

                self.log("Rx", connection=connection, status_code=status_code, body=response_body)

                response = v3io.dataplane.response.Response(request.output, status_code, headers, response_body)

                self._free_connections.put(connection, block=True)

                response.raise_for_status(request.raise_for_status or raise_for_status)

                return response

            except v3io.dataplane.response.HttpResponseError as response_error:
                self._logger.warn_with(f"Response error: {response_error}")
                raise response_error
            except BaseException as e:
                connection.close()
                connection = self._create_connection(self._host, self._ssl_context)

                if num_retries == 0:
                    self._logger.error_with(
                        "Error occurred while waiting for response and ran out of retries",
                        e=type(e),
                        e_msg=e,
                        response_body=response_body,
                        status_code=status_code,
                        headers=headers,
                    )
                    self._free_connections.put(connection, block=True)
                    raise e

                self._logger.debug_with(
                    "Error occurred while waiting for response – retrying",
                    retries_left=num_retries,
                    e=type(e),
                    e_msg=e,
                )

            num_retries -= 1
            is_retry = True

    def _send_request_on_connection(self, request, connection):
        request.transport.connection_used = connection

        path = request.encode_path()

        self.log(
            "Tx", connection=connection, method=request.method, path=path, headers=request.headers, body=request.body
        )
        starting_offset = 0
        is_body_seekable = request.body and hasattr(request.body, "seek") and hasattr(request.body, "tell")
        if is_body_seekable:
            starting_offset = request.body.tell()
        try:
            try:
                connection.request(request.method, path, request.body, request.headers)
            except self._send_request_exceptions as e:
                self._logger.debug_with(
                    "Disconnected while attempting to send. Recreating connection and retrying",
                    e=type(e),
                    e_msg=e,
                    connection=connection,
                )
                connection.close()
                if is_body_seekable:
                    # If the first connection fails, the pointer of the body might move at the size
                    # of the first connection blocksize.
                    # We need to reset the position of the pointer in order to send the whole file.
                    request.body.seek(starting_offset)
                connection = self._create_connection(self._host, self._ssl_context)
                request.transport.connection_used = connection
                connection.request(request.method, path, request.body, request.headers)
        except BaseException as e:
            self._logger.error_with(
                "Unhandled exception while sending request", e=type(e), e_msg=e, connection=connection
            )
            raise e

        return request

    def _create_connections(self, num_connections, host, ssl_context):
        for _ in range(num_connections):
            connection = self._create_connection(host, ssl_context)
            self._free_connections.put(connection, block=True)

    def _create_connection(self, host, ssl_context):
        if ssl_context is None:
            return http.client.HTTPConnection(host)

        return http.client.HTTPSConnection(host, context=ssl_context)

    def _parse_endpoint(self, endpoint):
        if endpoint.startswith("http://"):
            return endpoint[len("http://") :], None

        if endpoint.startswith("https://"):
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

            return endpoint[len("https://") :], ssl_context

        return endpoint, None

    def _get_status_and_headers_py2(self, response):
        return response.status, response.getheaders()

    def _get_status_and_headers_py3(self, response):
        return response.code, response.headers



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import v3io.dataplane.kv_cursor
import v3io.dataplane.model
import v3io.dataplane.output
import v3io.dataplane.request


class Model(v3io.dataplane.model.Model):
    def __init__(self, client):
        self._client = client
        self._access_key = client._access_key
        self._transport = client._transport

    async def head(self, container, path, access_key=None, raise_for_status=None):
        """Retrieves system attributes of object from a container.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        path (Required) : str
            The path of the object
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.

        Return Value
        ----------
        A `Response` object, whose `headers` is populated with system attributes of the object.
        """
        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_head_object,
            locals(),
        )

    async def get(self, container, path, access_key=None, raise_for_status=None, offset=None, num_bytes=None):
        """Retrieves an object from a container.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        path (Required) : str
            The path of the object
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        offset (Optional) : int
            A numeric offset into the object (in bytes). Defaults to 0
        num_bytes (Optional) : int
            Number of bytes to return. By default equal to len(object)-offset

        Return Value
        ----------
        A `Response` object, whose `body` is populated with the body of the object.
        """
        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_get_object,
            locals(),
        )

    async def put(self, container, path, access_key=None, raise_for_status=None, body=None, append=None):
        """Adds a new object to a container, or appends data to an existing object. The option to append data is
        extension to the S3 PUT Object capabilities

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        path (Required) : str
            The path of the object
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        body (Optional) : str
            The contents of the object
        append (Optional) : bool
            If True, the put appends the data to the end of the object. Defaults to False

        Return Value
        ----------
        A `Response` object
        """
        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_put_object,
            locals(),
        )

    async def delete(self, container, path, access_key=None, raise_for_status=None):
        """Deletes an object from a container.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        path (Required) : str
            The path of the object
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.

        Return Value
        ----------
        A `Response` object.
        """
        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_delete_object,
            locals(),
        )



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime
import sys

# used only n py2
BASE_DATETIME = datetime.datetime(1970, 1, 1)


def _get_timestamp_from_datetime_py3(dt):
    return dt.astimezone(datetime.timezone.utc).timestamp()


def _get_timestamp_from_datetime_py2(dt):
    return (dt - BASE_DATETIME).total_seconds()


# _get_timestamp_from_datetime is python version specific. resolve this once
if sys.version_info[0] >= 3:
    _get_timestamp_from_datetime = _get_timestamp_from_datetime_py3
else:
    _get_timestamp_from_datetime = _get_timestamp_from_datetime_py2


def encode(dt):
    timestamp = _get_timestamp_from_datetime(dt)

    # get integer and float part of nanoseconds
    seconds, nanoseconds_float = divmod(timestamp, 1)
    nanoseconds = int(nanoseconds_float * 1e9)

    return "{}:{}".format(int(seconds), nanoseconds)


def decode(encoded_dt):
    seconds_str, nanoseconds_str = encoded_dt.split(":")

    timestamp = int(seconds_str) + (int(nanoseconds_str) / 1e9)

    return datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import functools

import v3io.dataplane.transport


class Batch(object):
    def __init__(self, client):
        self._client = client
        self._encoded_requests = []
        self._inflight_requests = []
        self._transport_actions = v3io.dataplane.transport.Actions.encode_only
        self._transport = self._client._transport
        self.kv = lambda: None
        self.object = lambda: None
        self.stream = lambda: None
        self.container = lambda: None

        for client_call in [
            "get_containers",
            "get_container_contents",
            "get_object",
            "put_object",
            "delete_object",
            "put_item",
            "update_item",
            "get_item",
            "get_items",
            "create_stream",
            "delete_stream",
            "describe_stream",
            "seek_shard",
            "put_records",
            "get_records",
        ]:
            setattr(self, client_call, functools.partial(self._call_client, client_call))

        for model_name, model_call in [
            ("kv", "put"),
            ("kv", "get"),
            ("kv", "scan"),
            ("kv", "update"),
            ("kv", "delete"),
            ("object", "put"),
            ("object", "get"),
            ("object", "delete"),
            ("stream", "create"),
            ("stream", "update"),
            ("stream", "delete"),
            ("stream", "describe"),
            ("stream", "seek"),
            ("stream", "put_records"),
            ("stream", "get_records"),
            ("container", "get"),
            ("container", "list"),
        ]:
            setattr(getattr(self, model_name), model_call, functools.partial(self._call_model, model_name, model_call))

    def _call_client(self, name, *args, **kw_args):
        kw_args["transport_actions"] = self._transport_actions
        request = getattr(self._client, name)(*args, **kw_args)

        self._encoded_requests.append(request)

    def _call_model(self, model_name, model_call, *args, **kw_args):
        kw_args["transport_actions"] = self._transport_actions

        # get the model (kv, object, ...)
        model = getattr(self._client, model_name)

        # do the request on it
        request = getattr(model, model_call)(*args, **kw_args)

        # shove to encoded requests
        self._encoded_requests.append(request)

    def wait(self, raise_for_status=None):
        try:
            return self._wait(raise_for_status)

        # if an exception is raised, clean up everything
        except Exception as e:
            self._inflight_requests = []
            self._encoded_requests = []
            self._transport.restart()

            raise e

    def _wait(self, raise_for_status=None):
        responses = []

        # while we can send requests - send them
        while self._encoded_requests and len(self._inflight_requests) < self._transport.max_connections:
            # send the request
            request = self._transport.send_request(self._encoded_requests.pop(0))

            # add to inflight requests
            self._inflight_requests.append(request)

        # start creating responses
        while self._inflight_requests:
            # get an inflight request
            inflight_request = self._inflight_requests.pop(0)

            # wait for the response of the request
            response = self._transport.wait_response(inflight_request, raise_for_status)

            # add to responses
            responses.append(response)

            # if there's a pending request, send it on the connection that we just read from
            if self._encoded_requests:
                # send the request
                request = self._transport.send_request(self._encoded_requests.pop(0))

                # add to inflight requests
                self._inflight_requests.append(request)

        return responses



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import requests

import v3io.dataplane.request
import v3io.dataplane.response

from . import abstract


class Transport(abstract.Transport):
    def __init__(self, logger, endpoint=None, max_connections=None, timeout=None, verbosity=None):
        super(Transport, self).__init__(logger, endpoint, max_connections, timeout, verbosity)
        self._next_connection_pool = 0
        self._session = requests.Session()

    def requires_access_key(self):
        return True

    def close(self):
        self._session.close()

    def send_request(self, request):
        path = request.encode_path()

        # call the encoder to get the response
        http_response = self._http_request(request.method, path, request.headers, request.body)

        # set http response
        request.transport.http_response = http_response

        return request

    def wait_response(self, request, raise_for_status=None, num_retries=1):
        # create a response
        response = v3io.dataplane.response.Response(
            request.output,
            request.transport.http_response.status_code,
            request.headers,
            request.transport.http_response.content,
        )

        # enforce raise for status
        response.raise_for_status(request.raise_for_status or raise_for_status)

        return response

    def _http_request(self, method, path, headers=None, body=None):
        self.log("Tx", method=method, path=path, headers=headers, body=body)

        response = self._session.request(
            method, self._endpoint + path, headers=headers, data=body, timeout=self._timeout, verify=False
        )

        self.log("Rx", status_code=response.status_code, headers=response.headers, body=response.text)

        return response



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .verifier import Transport  # noqa: F401


class RaiseForStatus(object):
    never = "never"
    always = "always"


class Actions(object):
    encode_only = 1
    encode_and_send = 2
    send_and_receive = 3



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

import v3io.aio.dataplane.kv_cursor
import v3io.dataplane.model
import v3io.dataplane.output
import v3io.dataplane.request


class Model(v3io.dataplane.model.Model):
    def __init__(self, client):
        self._client = client
        self._access_key = client._access_key
        self._transport = client._transport

    def new_cursor(
        self,
        container,
        table_path,
        table_name=None,
        access_key=None,
        raise_for_status=None,
        attribute_names="*",
        filter_expression=None,
        marker=None,
        sharding_key=None,
        limit=None,
        segment=None,
        total_segments=None,
        sort_key_range_start=None,
        sort_key_range_end=None,
    ):
        return v3io.aio.dataplane.kv_cursor.Cursor(
            self._client,
            container,
            access_key or self._access_key,
            table_path,
            table_name,
            raise_for_status,
            attribute_names,
            filter_expression,
            marker,
            sharding_key,
            limit,
            segment,
            total_segments,
            sort_key_range_start,
            sort_key_range_end,
        )

    async def put(self, container, table_path, key, attributes, access_key=None, raise_for_status=None, condition=None):
        """Creates an item with the provided attributes. If an item with the same name (primary key) already exists in
        the specified table, the existing item is completely overwritten (replaced with a new item). If the item or
        table do not exist, the operation creates them.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/nosql-web-api/putitem/.

        Notes:
        1. To provide arrays, pass either a list of integers ([1, 2, 3]), a list of floats ([1.0, 2.0, 3.0]) an
           array.array with a typecode of either 'l' (integer) or 'd' (float). The response will always either be
           a list of integers or a list of floats (never an array.array)
        2. To provide a timestamp, pass a datetime.datetime. Whatever the timezone, it will be stored as UTC and
           a UTC datetime will be retreived when read

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        table_path (Required) : str
            The full path of the table
        key (Required) : str
            The item key name
        attributes (Required) : dict
            The item to add - an object containing zero or more attributes.
            For example:
                {
                    'age': 42,
                    'feature': 'mustache'
                }

        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        condition (Optional) : str
            A Boolean condition expression that defines a conditional logic for executing the put-item operation.

        Return Value
        ----------
        A `Response` object.
        """

        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_put_item,
            locals(),
        )

    async def update(
        self,
        container,
        table_path,
        key,
        access_key=None,
        raise_for_status=None,
        attributes=None,
        expression=None,
        condition=None,
        update_mode=None,
        alternate_expression=None,
    ):
        """Updates the attributes of a table item. If the specified item or table don't exist,
        the operation creates them.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/nosql-web-api/updateitem/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        table_path (Required) : str
            The full path of the table
        key (Required) : str
            The item key name
        attributes (Required) : dict
            The item to update - an object containing zero or more attributes.
            For example:
                {
                    'age': 42,
                    'feature': 'mustache'
                }
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        expression (Optional) : str
            An update expression that specifies the changes to make to the item's attributes.
        condition (Optional) : str
            A Boolean condition expression that defines a conditional logic for executing the put-item operation.
            See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/nosql-web-api/putitem/.
        update_mode (Optional) : str
            CreateOrReplaceAttributes (default): Creates or replaces attributes
        alternate_expression (Optional) : str
            An alternate update expression that specifies the changes to make to the item's attributes when a
            condition expression, defined in the ConditionExpression request parameter, evaluates to false;
            (i.e., this parameter defines the else clause of a conditional if-then-else update expression).
            See Update Expression for syntax details and examples. When the alternate update expression is executed,
            it's evaluated against the table item to be updated, if it exists. If the item doesn't exist, the update
            creates it (as well as the parent table if it doesn't exist). See also the UpdateExpression notes, which
            apply to the alternate update expression as well.

        Return Value
        ----------
        A `Responses` object.
        """
        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_update_item,
            locals(),
        )

    async def get(self, container, table_path, key, access_key=None, raise_for_status=None, attribute_names="*"):
        """Retrieves the requested attributes of a table item.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/nosql-web-api/getitem/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        table_path (Required) : str
            The full path of the table
        key (Required) : str
            The item key name
        attribute_names (Required) : []str or '*'
            A list of attribute names to get, or '*' which will retreive all attributes
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.

        Return Value
        ----------
        A `Response` object, whose `output` is `GetItemOutput`.
        """
        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_get_item,
            locals(),
            v3io.dataplane.output.GetItemOutput,
        )

    async def scan(
        self,
        container,
        table_path,
        table_name=None,
        access_key=None,
        raise_for_status=None,
        attribute_names="*",
        filter_expression=None,
        marker=None,
        sharding_key=None,
        limit=None,
        segment=None,
        total_segments=None,
        sort_key_range_start=None,
        sort_key_range_end=None,
    ):
        """Retrieves (reads) attributes of multiple items in a table or in a data container's root directory,
        according to the specified criteria.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/nosql-web-api/getitems/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        table_path (Required) : str
            The full path of the table
        table_name (Optional) : str
            The name of the table. Allows for specifying all tables with '/*'. If used, table_path must be an empty
            string.
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        attribute_names (Optional) : []str or '*'
            A list of attribute names to get, or '*' which will retreive all attributes
        filter_expression (Optional) : str
            A filter expression that restricts the items to retrieve. Only items that match the filter criteria
            are returned.
            See https://www.iguazio.com/docs/latest-release/data-layer/reference/expressions/condition-expression/#filter-expression.
        marker (Optional) : str
            An opaque identifier that was returned in the NextMarker element of a response to a previous GetItems
            request that did not return all the requested items. This marker identifies the location in the table
            from which to start searching for the remaining requested items. See Partial Response and the description
            of the NextMarker response element.
        sharding_key (Optional) : str
            The maximal sorting-key value of the items to get by using a range scan. The sorting-key value is the
            part to the right of the leftmost period in a compound primary-key value (item name). This parameter is
            applicable only together with the ShardingKey request parameter. The scan will return all items with the
            specified sharding-key value whose sorting-key values are greater than or equal to (>=) than the value of
            the SortKeyRangeStart parameter (if set) and less than (<) the value of the SortKeyRangeEnd parameter.
        limit (Optional) : int
            The maximum number of items to return within the response (i.e., the maximum number of elements in the
            response object's Items array).
        segment (Optional) : str
            The ID of a specific table segment to scan - 0 to one less than TotalSegment
        total_segments (Optional) : str
            The number of segments into which to divide the table scan - 1 to 1024. See Parallel Scan.
            The segments are assigned sequential IDs starting with 0.
        sort_key_range_start (Optional) : str
            The minimal sorting-key value of the items to get by using a range scan. The sorting-key value is the part
            to the right of the leftmost period in a compound primary-key value (item name). This parameter is
            applicable only together with the ShardingKey request parameter. The scan will return all items with
            the specified sharding-key value whose sorting-key values are greater than or equal to (>=) the value of
            the SortKeyRangeStart parameter and less than (<) the value of the SortKeyRangeEnd parameter (if set).
        sort_key_range_end (Optional) : str
            The maximal sorting-key value of the items to get by using a range scan. The sorting-key value is the part
             to the right of the leftmost period in a compound primary-key value (item name). This parameter is
             applicable only together with the ShardingKey request parameter. The scan will return all items with
             the specified sharding-key value whose sorting-key values are greater than or equal to (>=) than the
             value of the SortKeyRangeStart parameter (if set) and less than (<) the value of the SortKeyRangeEnd
             parameter.

        Return Value
        ----------
        A `Response` object, whose `output` is `GetItemsOutput`.
        """  # noqa
        table_path = self._ensure_path_ends_with_slash(table_path)

        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_get_items,
            locals(),
            v3io.dataplane.output.GetItemsOutput,
        )

    async def delete(self, container, table_path, key, access_key=None, raise_for_status=None, transport_actions=None):
        """Deletes an item.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        table_path (Required) : str
            The full path of the table
        key (Required) : str
            The item key name
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.

        Return Value
        ----------
        A `Response` object.
        """
        return self._client.delete_object(
            container, os.path.join(table_path, key), access_key, raise_for_status, transport_actions
        )

    async def create_schema(self, container, table_path, access_key=None, raise_for_status=None, key=None, fields=None):
        """Creates a KV schema file

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        table_path (Required) : str
            The full path of the table
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        key (Required) : str
            The key field name
        fields (Required) : list of dicts
            A dictionary of fields, where each item has:
            - name (string)
            - type (string - one of string, double, long)
            - nullable (boolean)

            Example: [
                {
                  'name': 'my_field',
                  'type': 'string',
                  'nullable': False
                },
                {
                  'name': 'who',
                  'type': 'string',
                  "nullable": True
                }
            ]

        Return Value
        ----------
        A `Response` object
        """
        put_object_args = locals()
        put_object_args["path"] = os.path.join(put_object_args["table_path"], ".#schema")
        put_object_args["offset"] = 0
        put_object_args["append"] = None
        put_object_args["body"] = self._client._get_schema_contents(key, fields)
        del put_object_args["key"]
        del put_object_args["fields"]

        return await self._transport.request(
            container,
            access_key or self._access_key,
            raise_for_status,
            v3io.dataplane.request.encode_put_object,
            put_object_args,
        )



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64

import future.utils

import v3io.dataplane.kv_array
import v3io.dataplane.kv_timestamp


class Output(object):
    def _decode_typed_attributes(self, typed_attributes):
        decoded_attributes = {}

        for attribute_key, typed_attribute_value in future.utils.viewitems(typed_attributes):
            for attribute_type, attribute_value in future.utils.viewitems(typed_attribute_value):
                if attribute_type == "N":
                    try:
                        decoded_attribute = int(attribute_value)
                    except ValueError:
                        decoded_attribute = float(attribute_value)
                elif attribute_type == "B":
                    decoded_attribute = base64.b64decode(attribute_value)

                    # try to decode as an array
                    try:
                        decoded_attribute = v3io.dataplane.kv_array.decode(decoded_attribute)
                    except BaseException:
                        pass

                elif attribute_type == "S":
                    if type(attribute_value) in [float, int]:
                        decoded_attribute = str(attribute_value)
                    else:
                        decoded_attribute = attribute_value

                elif attribute_type == "TS":
                    decoded_attribute = v3io.dataplane.kv_timestamp.decode(attribute_value)
                else:
                    decoded_attribute = attribute_value

                decoded_attributes[attribute_key] = decoded_attribute

        return decoded_attributes


#
# Containers
#


class Container(object):
    def __init__(self, name, creation_date, identifier):
        self.name = name
        self.creation_date = creation_date
        self.identifier = identifier


class GetContainersOutput(Output):
    def __init__(self, root):
        # got an error code
        if isinstance(root, dict):
            self.error = root
            return

        self.containers = []

        for bucket in root.find("Buckets"):
            self.containers.append(
                Container(
                    name=bucket.find("Name").text,
                    creation_date=bucket.find("CreationDate").text,
                    identifier=int(bucket.find("Id").text),
                )
            )


class ContainerContent(object):
    def __init__(self, child):
        # got an error code
        if isinstance(child, dict):
            self.error = child
            return

        for child_key, attribute_name, kind in [
            ("Key", "key", str),
            ("Size", "size", int),
            ("LastSequenceID", "last_sequence_id", int),
            ("LastModified", "last_modified", str),
            ("Mode", "mode", str),
            ("AccessTime", "access_time", str),
            ("CreatingTime", "creating_time", str),
            ("GID", "gid", str),
            ("UID", "uid", str),
            ("InodeNumber", "inode_number", int),
        ]:
            child_value = child.find(child_key)
            if child_value is not None:
                setattr(self, attribute_name, kind(child_value.text))


class ContainerCommonPrefix(object):
    def __init__(self, child):
        # got an error code
        if isinstance(child, dict):
            self.error = child
            return

        for child_key, attribute_name, kind in [
            ("Prefix", "prefix", str),
            ("LastModified", "last_modified", str),
            ("AccessTime", "access_time", str),
            ("CreatingTime", "creating_time", str),
            ("Mode", "mode", str),
            ("GID", "gid", str),
            ("UID", "uid", str),
            ("InodeNumber", "inode_number", int),
        ]:
            child_value = child.find(child_key)
            if child_value is not None:
                setattr(self, attribute_name, kind(child_value.text))


class GetContainerContentsOutput(Output):
    def __init__(self, root):
        # got an error code
        if isinstance(root, dict):
            self.error = root
            return

        self.name = root.find("Name").text
        self.next_marker = root.find("NextMarker").text
        self.max_keys = root.find("MaxKeys").text
        self.is_truncated = root.find("IsTruncated").text
        self.contents = []
        self.common_prefixes = []

        contents_children = root.findall("Contents")
        common_prefixes_children = root.findall("CommonPrefixes")

        if contents_children is not None:
            for content_child in contents_children:
                self.contents.append(ContainerContent(content_child))

        if common_prefixes_children is not None:
            for common_prefix_child in common_prefixes_children:
                self.common_prefixes.append(ContainerCommonPrefix(common_prefix_child))


#
# KV/EMD
#


class GetItemOutput(Output):
    def __init__(self, decoded_body):
        self.item = self._decode_typed_attributes(decoded_body.get("Item", {}))


class GetItemsOutput(Output):
    def __init__(self, decoded_body):
        self.last = decoded_body.get("LastItemIncluded") == "TRUE"
        self.next_marker = decoded_body.get("NextMarker")
        self.items = []

        for item in decoded_body.get("Items", []):
            self.items.append(self._decode_typed_attributes(item))


#
# Stream
#


class DescribeStreamOutput(Output):
    def __init__(self, decoded_body):
        self.shard_count = decoded_body.get("ShardCount")
        self.retention_period_hours = decoded_body.get("RetentionPeriodHours")


class SeekShardOutput(Output):
    def __init__(self, decoded_body):
        self.location = decoded_body.get("Location")


class PutRecordsResult(Output):
    def __init__(self, decoded_body):
        self.sequence_number = decoded_body.get("SequenceNumber")
        self.shard_id = decoded_body.get("ShardId")
        self.error_code = decoded_body.get("ErrorCode")
        self.error_message = decoded_body.get("ErrorMessage")


class PutRecordsOutput(Output):
    def __init__(self, decoded_body):
        self.failed_record_count = decoded_body.get("FailedRecordCount")
        self.records = []

        for record in decoded_body.get("Records", []):
            self.records.append(PutRecordsResult(record))


class GetRecordsResult(Output):
    def __init__(self, decoded_body):
        self.arrival_time_sec = decoded_body.get("ArrivalTimeSec")
        self.arrival_time_nsec = decoded_body.get("ArrivalTimeNSec")
        self.sequence_number = decoded_body.get("SequenceNumber")
        self.client_info = self._from_base64(decoded_body.get("ClientInfo"))
        self.partition_key = decoded_body.get("PartitionKey")
        self.data = self._from_base64(decoded_body.get("Data"))

    @staticmethod
    def _from_base64(value):
        if value is None:
            return None

        return base64.b64decode(value)


class GetRecordsOutput(Output):
    def __init__(self, decoded_body):
        self.next_location = decoded_body.get("NextLocation")
        self.msec_behind_latest = decoded_body.get("MSecBehindLatest")
        self.records_behind_latest = decoded_body.get("RecordsBehindLatest")
        self.records = []

        for record in decoded_body.get("Records"):
            self.records.append(GetRecordsResult(record))



import itertools
import mmap
import os
import random
import tempfile
from http.client import CannotSendRequest

import pytest

import v3io.dataplane.client
import v3io.dataplane.transport.httpclient
from v3io.dataplane.object import Model
from v3io.dataplane.response import HttpResponseError


class MockResponse:
    def __init__(self, code=None, headers=None):
        self.code = code or 200
        self.headers = headers or {}

    def read(self):
        return "mock response"

    def getheaders(self):
        return self.headers


class ATestException(Exception):
    pass


class MockConnection:
    def __init__(self, fail_request_after=None, fail_getresponse_after=None, exception_class=ATestException):
        self.times_closed = 0
        self.times_request = 0
        self.times_got_response = 0

        self.fail_request_after = fail_request_after
        self.fail_getresponse_after = fail_getresponse_after
        self.exception_class = exception_class

    def request(self, method, path, body, headers):
        self.times_request += 1
        data = b""
        if hasattr(body, "read"):
            data = body.read()
        if self.fail_request_after is not None and self.times_request > self.fail_request_after:
            raise self.exception_class(f"Failing request number {self.times_request}")
        if method == "PUT":
            if not os.path.exists(path):
                return
            with open(path, "wb") as file:
                file.write(data)

    def getresponse(self):
        self.times_got_response += 1
        if self.fail_getresponse_after is not None and self.times_got_response > self.fail_getresponse_after:
            raise self.exception_class(f"Failing response number {self.times_got_response}")
        return MockResponse()

    def raise_for_status(self, expected_statuses=None):
        pass

    def close(self):
        self.times_closed += 1


class MockTransport(v3io.dataplane.transport.httpclient.Transport):
    def __init__(self, *args, connection_options=None, reset_after_create_connections=False, **kwargs):
        self.mock_connections = []
        self.connection_options = connection_options or {}
        self.reset_after_create_connections = reset_after_create_connections
        super().__init__(*args, **kwargs)

    def _create_connection(self, host, ssl_context):
        conn = MockConnection(**self.connection_options)
        self.mock_connections.append(conn)
        return conn

    def _create_connections(self, num_connections, host, ssl_context):
        super()._create_connections(num_connections=num_connections, host=host, ssl_context=ssl_context)
        if self.reset_after_create_connections:
            self.connection_options = {}


def test_first_connection_failure():
    connection_options = {"fail_request_after": 0, "exception_class": CannotSendRequest}
    client = v3io.dataplane.Client()

    mock_transport = MockTransport(
        client._logger, connection_options=connection_options, reset_after_create_connections=True
    )
    client._transport = mock_transport
    size = 1024
    data = random.Random(0).randbytes(size)
    with mmap.mmap(-1, size) as mmap_obj, tempfile.NamedTemporaryFile(mode="w+b", delete=False) as temp_file:
        mmap_obj.write(data)
        mmap_obj.seek(0)
        components = temp_file.name.split("/")
        container = components[1]  # Index 0 will be an empty string due to leading '/'
        path = "/" + "/".join(components[2:])
        client.put_object(
            container=container, path=path, body=mmap_obj, raise_for_status=v3io.dataplane.RaiseForStatus.never
        )
        temp_file.seek(0)
        assert temp_file.read() == data, "Binary data read back differs from original data"


def test_connection_creation_and_close():
    client = v3io.dataplane.Client()
    mock_transport = MockTransport(client._logger)
    client._transport = mock_transport
    for _ in range(20):
        client.get_object("mycontainer", "mypath")
    assert len(mock_transport.mock_connections) == 8
    for conn in mock_transport.mock_connections:
        assert conn.times_closed == 0
    client.close()
    assert len(mock_transport.mock_connections) == 8
    for conn in mock_transport.mock_connections:
        assert conn.times_closed == 1


@pytest.mark.parametrize(["fail_request_after", "fail_getresponse_after"], itertools.permutations(range(3), 2))
def test_connection_closed_on_error(fail_request_after, fail_getresponse_after):
    client = v3io.dataplane.Client()
    connection_options = {
        "fail_request_after": fail_request_after,
        "fail_getresponse_after": fail_getresponse_after,
    }
    mock_transport = MockTransport(client._logger, connection_options=connection_options)
    client._transport = mock_transport
    for _ in range(20):
        try:
            client.get_object("mycontainer", "mypath")
        except ATestException:
            pass
    num_connections_still_open = 0
    for conn in mock_transport.mock_connections:
        assert conn.times_closed == 0 or conn.times_closed == 1
        if conn.times_closed == 0:
            num_connections_still_open += 1
    assert num_connections_still_open == 8
    client.close()
    for conn in mock_transport.mock_connections:
        assert conn.times_closed == 1


def test_error_on_use_after_close():
    client = v3io.dataplane.Client()
    client.object.get("doesntexist", "doesntexist", raise_for_status=v3io.dataplane.RaiseForStatus.never)
    client.close()
    with pytest.raises(RuntimeError):
        client.object.get("doesntexist", "doesntexist", raise_for_status=v3io.dataplane.RaiseForStatus.never)


@pytest.mark.parametrize("object_function", [Model.get, Model.put, Model.delete])
def test_raise_http_response_error(object_function):
    client = v3io.dataplane.Client(
        transport_kind="httpclient",
    )
    with pytest.raises(HttpResponseError, match="Container not found") as response_error:
        object_function(client.object, "not-exists", "path/to/object")
    assert response_error.value.status_code == 404



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
def url_join(base, *parts):
    result = base

    if result[0] != "/":
        result = "/" + base

    for part in parts:
        if part[0] != "/":
            result += "/" + part
        else:
            result += part

    return result



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import array
import datetime
import os
import unittest

import future.utils

import v3io.aio.dataplane
import v3io.dataplane


class Test(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self._client = v3io.aio.dataplane.Client(logger_verbosity="DEBUG", transport_verbosity="DEBUG")

        self._container = "bigdata"
        self._test_parent_dir = os.environ.get("MATRIX_PYTHON_VERSION")
        if self._test_parent_dir is None:
            self._test_parent_dir = ""

    async def asyncTearDown(self):
        await self._client.close()

    async def _delete_dir(self, path):
        response = await self._client.container.list(
            container=self._container, path=path, raise_for_status=v3io.aio.dataplane.RaiseForStatus.never
        )

        if response.status_code == 404:
            return

        if response.status_code != 200:
            raise RuntimeError(response.body)

        for content in response.output.contents:
            await self._client.object.delete(container=self._container, path=content.key)

        for common_prefixes in response.output.common_prefixes:
            await self._client.object.delete(container=self._container, path=common_prefixes.prefix)


class TestContainer(Test):
    async def asyncSetUp(self):
        await super(TestContainer, self).asyncSetUp()
        self._path = os.path.join(self._test_parent_dir, "v3io-py-test-container")

        # clean up
        await self._delete_dir(self._path)

    async def test_get_container_contents_invalid_path(self):
        response = await self._client.container.list(
            container=self._container, path="/no-such-path", raise_for_status=v3io.aio.dataplane.RaiseForStatus.never
        )
        self.assertEqual(404, response.status_code)
        self.assertIn("No such file", str(response.body))

    async def test_get_container_contents(self):
        body = "If you cannot do great things, do small things in a great way."

        for object_index in range(5):
            await self._client.object.put(
                container=self._container,
                path=os.path.join(self._path, "object-{0}.txt".format(object_index)),
                body=body,
            )

        for object_index in range(5):
            await self._client.object.put(
                container=self._container, path=os.path.join(self._path, "dir-{0}/".format(object_index))
            )

        response = await self._client.container.list(
            container=self._container, path=self._path, get_all_attributes=True, directories_only=True
        )
        self.assertEqual(0, len(response.output.contents))
        self.assertNotEqual(0, len(response.output.common_prefixes))

        response = await self._client.container.list(
            container=self._container, path=self._path, get_all_attributes=True
        )
        self.assertNotEqual(0, len(response.output.contents))
        self.assertNotEqual(0, len(response.output.common_prefixes))

        # clean up
        await self._delete_dir(self._path)


class TestStream(Test):
    async def asyncSetUp(self):
        await super(TestStream, self).asyncSetUp()

        self._path = os.path.join(self._test_parent_dir, "v3io-py-test-stream")

        # clean up
        await self._client.stream.delete(
            container=self._container, stream_path=self._path, raise_for_status=[200, 204, 404]
        )

    async def test_delete_stream_with_cg(self):
        num_shards = 8

        # check that the stream doesn't exist
        self.assertFalse(await self._stream_exists())

        # create a stream
        await self._client.stream.create(container=self._container, stream_path=self._path, shard_count=num_shards)

        # write data to all shards so there are files
        for shard_id in range(num_shards):
            await self._client.stream.put_records(
                container=self._container,
                stream_path=self._path,
                records=[{"shard_id": shard_id, "data": "data for shard {}".format(shard_id)}],
            )

        # write several "consumer group state" files
        for cg_id in range(3):
            await self._client.object.put(
                container=self._container, path=os.path.join(self._path, "cg{}-state.json".format(cg_id))
            )

        # check that the stream doesn't exist
        self.assertTrue(await self._stream_exists())

        # delete the stream
        await self._client.stream.delete(container=self._container, stream_path=self._path)

        # check that the stream doesn't exist
        self.assertFalse(await self._stream_exists())

    async def test_stream(self):
        # create a stream w/8 shards
        await self._client.stream.create(container=self._container, stream_path=self._path, shard_count=8)

        records = [
            {"shard_id": 1, "data": "first shard record #1"},
            {"shard_id": 1, "data": "first shard record #2", "client_info": bytearray(b"some info")},
            {"shard_id": 10, "data": "invalid shard record #1"},
            {"shard_id": 2, "data": "second shard record #1"},
            {"data": "some shard record #1"},
        ]

        response = await self._client.stream.put_records(
            container=self._container, stream_path=self._path, records=records
        )
        self.assertEqual(1, response.output.failed_record_count)

        for response_record_index, response_record in enumerate(response.output.records):
            if response_record_index == 2:
                self.assertIsNotNone(response_record.error_code)
            else:
                self.assertIsNone(response_record.error_code)

        response = await self._client.stream.seek(
            container=self._container, stream_path=self._path, shard_id=1, seek_type="EARLIEST"
        )

        self.assertNotEqual("", response.output.location)

        response = await self._client.stream.get_records(
            container=self._container, stream_path=self._path, shard_id=1, location=response.output.location
        )

        self.assertEqual(2, len(response.output.records))
        self.assertEqual(records[0]["data"], response.output.records[0].data.decode("utf-8"))
        self.assertEqual(records[1]["data"], response.output.records[1].data.decode("utf-8"))
        self.assertEqual(records[1]["client_info"], response.output.records[1].client_info)

        # update the stream by adding 8 shards to it
        await self._client.stream.update(container=self._container, stream_path=self._path, shard_count=16)

        records = [
            {"shard_id": 10, "data": "Now valid shard record #1"},
        ]

        response = await self._client.stream.put_records(
            container=self._container, stream_path=self._path, records=records
        )

        self.assertEqual(0, response.output.failed_record_count)

        await self._client.stream.delete(container=self._container, stream_path=self._path)

    async def _stream_exists(self):
        response = await self._client.stream.describe(
            container=self._container, stream_path=self._path, raise_for_status=v3io.aio.dataplane.RaiseForStatus.never
        )
        return response.status_code == 200


class TestObject(Test):
    async def asyncSetUp(self):
        await super(TestObject, self).asyncSetUp()

        self._object_dir = os.path.join(self._test_parent_dir, "v3io-py-test-object")
        self._object_path = self._object_dir + "/obj ect.txt"

        # clean up
        await self._delete_dir(self._object_dir)

    async def test_object(self):
        contents = "vegans are better than everyone"

        response = await self._client.object.get(
            container=self._container, path=self._object_path, raise_for_status=v3io.aio.dataplane.RaiseForStatus.never
        )

        self.assertEqual(404, response.status_code)

        # put contents to some object
        await self._client.object.put(container=self._container, path=self._object_path, body=contents)

        # get the contents
        response = await self._client.object.get(container=self._container, path=self._object_path)

        if not isinstance(response.body, str):
            response.body = response.body.decode("utf-8")

        self.assertEqual(response.body, contents)

        # get the head of the object
        response = await self._client.object.head(container=self._container, path=self._object_path)

        self.assertIn(("Content-Length", str(len(contents))), response.headers.items())

        # get the head of the dir-object
        response = await self._client.object.head(container=self._container, path=self._object_dir)

        self.assertIn(("Content-Length", str(0)), response.headers.items())

        # delete the object
        await self._client.object.delete(container=self._container, path=self._object_path)

        # get again
        response = await self._client.object.get(
            container=self._container, path=self._object_path, raise_for_status=v3io.aio.dataplane.RaiseForStatus.never
        )

        self.assertEqual(404, response.status_code)

    async def test_append(self):
        contents = [
            "First part",
            "Second part",
            "Third part",
        ]

        # put the contents into the object
        for content in contents:
            await self._client.object.put(container=self._container, path=self._object_path, body=content, append=True)

        # get the contents
        response = await self._client.object.get(container=self._container, path=self._object_path)

        self.assertEqual(response.body.decode("utf-8"), "".join(contents))

    async def test_get_offset(self):
        await self._client.object.put(container=self._container, path=self._object_path, body="1234567890")

        # get the contents without limit
        response = await self._client.object.get(container=self._container, path=self._object_path, offset=4)

        self.assertEqual(response.body.decode("utf-8"), "567890")

        # get the contents with limit
        response = await self._client.object.get(
            container=self._container, path=self._object_path, offset=4, num_bytes=3
        )

        self.assertEqual(response.body.decode("utf-8"), "567")


# class TestSchema(Test):
#     async def asyncSetUp(self):
#         await super(TestSchema, self).asyncSetUp()
#
#         self._schema_dir = "/v3io-py-test-schemaa"
#         self._schema_path = os.path.join(self._schema_dir, ".%23schema")
#
#         # clean up
#         await self._delete_dir(self._schema_dir)
#
#     async def test_create_schema(self):
#         await self._client.kv.create_schema(
#             container=self._container,
#             table_path=self._schema_dir,
#             key="key_field",
#             fields=[
#                 {"name": "key_field", "type": "string", "nullable": False},
#                 {"name": "data_field_0", "type": "long", "nullable": True},
#                 {"name": "data_field_1", "type": "double", "nullable": True},
#             ],
#         )
#
#         # write to test the values in the UI (requires breaking afterwards)
#         items = {
#             "a": {"data_field_0": 30, "data_field_1": 100},
#             "b": {"data_field_0": 300, "data_field_1": 1000},
#             "c": {"data_field_0": 3000, "data_field_1": 10000},
#         }
#
#         for item_key, item_attributes in future.utils.viewitems(items):
#             await self._client.kv.put(
#                 container=self._container, table_path=self._schema_dir, key=item_key, attributes=item_attributes
#             )
#
#         # verify the scehma
#         response = await self._client.object.get(
#             container=self.container,
#             path=self._schema_path,
#             raise_for_status=v3io.aio.dataplane.RaiseForStatus.never,
#         )
#
#         expected = {
#             "hashingBucketNum": 0,
#             "key": "key_field",
#             "fields": [
#                 {"name": "key_field", "type": "string", "nullable": False},
#                 {"name": "data_field_0", "type": "long", "nullable": True},
#                 {"name": "data_field_1", "type": "double", "nullable": True},
#             ],
#         }
#
#         self.assertEqual(
#             json.loads(response.body.decode("utf-8")),
#             expected,
#         )


class TestKv(Test):
    async def asyncSetUp(self):
        await super(TestKv, self).asyncSetUp()

        self._path = os.path.join(self._test_parent_dir, "some_dir/v3io-py-test-emd")
        await self._delete_dir(self._path)

    async def test_kv_array(self):
        item_key = "item_with_arrays"
        item = {
            "list_with_ints": [1, 2, 3],
            "list_with_floats": [10.25, 20.25, 30.25],
        }

        await self._client.kv.put(container=self._container, table_path=self._path, key=item_key, attributes=item)

        for attribute_name in item.keys():
            await self._client.kv.update(
                container=self._container,
                table_path=self._path,
                key=item_key,
                expression=f"{attribute_name}[1]={attribute_name}[1]*2",
            )

        # get the item
        response = await self._client.kv.get(container=self._container, table_path=self._path, key=item_key)

        for attribute_name in item.keys():
            self.assertEqual(response.output.item[attribute_name][1], item[attribute_name][1] * 2)

    async def test_kv_values(self):
        def _get_int_array():
            int_array = array.array("l")
            for value in range(10):
                int_array.append(value)

            return int_array

        def _get_float_array():
            float_array = array.array("d")
            for value in range(10):
                float_array.append(value)

            return float_array

        item_key = "bob"
        item = {
            item_key: {
                "age": 42,
                "pi": 3.14,
                "feature_str": "mustache",
                "feature_unicode": "mustache",
                "numeric_str": "1",
                "unicode": "\xd7\xa9\xd7\x9c\xd7\x95\xd7\x9d",
                "male": True,
                "happy": False,
                "blob": b"+AFymWFzAL/LUOiU2huiADbugMH0AARATEO1",
                "list_with_ints": [1, 2, 3],
                "list_with_floats": [10.5, 20.5, 30.5],
                "array_with_ints": _get_int_array(),
                "array_with_floats": _get_float_array(),
                "now": datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc),
            }
        }

        await self._client.kv.put(
            container=self._container, table_path=self._path, key=item_key, attributes=item[item_key]
        )

        response = await self._client.kv.get(container=self._container, table_path=self._path, key=item_key)

        self.assertEqual(len(item[item_key].keys()), len(response.output.item.keys()))

        for key, value in response.output.item.items():
            self._compare_item_values(item[item_key][key], value)

        for key in item[item_key]:
            self._compare_item_types(item[item_key][key], response.output.item[key])

    async def test_kv(self):
        items = {
            "bob": {"age": 42, "feature": "mustache"},
            "linda": {"age": 41, "feature": "singing"},
            "louise": {"age": 9, "feature": "bunny ears"},
            "tina": {"age": 14, "feature": "butts"},
        }

        for item_key, item_attributes in future.utils.viewitems(items):
            await self._client.kv.put(
                container=self._container, table_path=self._path, key=item_key, attributes=item_attributes
            )

        await self._verify_items(self._path, items)

        await self._client.kv.update(
            container=self._container,
            table_path=self._path,
            key="louise",
            attributes={"height": 130, "quip": "i can smell fear on you"},
        )

        response = await self._client.kv.get(
            container=self._container,
            table_path=self._path,
            key="louise",
            attribute_names=["__size", "age", "quip", "height"],
        )

        self.assertEqual(0, response.output.item["__size"])
        self.assertEqual(9, response.output.item["age"])
        self.assertEqual("i can smell fear on you", response.output.item["quip"])
        self.assertEqual(130, response.output.item["height"])

        # get items with filter expression
        response = await self._client.kv.scan(
            container=self._container, table_path=self._path, filter_expression="feature == 'singing'"
        )
        self.assertEqual(1, len(response.output.items))

        # get items with segment / total_segments
        total_segments = 4
        total_items = []

        for segment in range(total_segments):
            received_items = await self._client.kv.new_cursor(
                container=self._container, table_path=self._path, segment=segment, total_segments=total_segments
            ).all()
            total_items.append(received_items)

        self.assertEqual(4, len(total_items))

        received_items = await self._client.kv.new_cursor(
            container=self._container,
            table_path=self._path,
            attribute_names=["age", "feature"],
            filter_expression="age > 15",
        ).all()

        self.assertEqual(2, len(received_items))
        for item in received_items:
            self.assertLess(15, item["age"])

        received_items = await self._client.kv.new_cursor(
            container=self._container,
            table_path="",
            table_name=self._path,
            attribute_names=["age", "feature"],
            filter_expression="age > 15",
        ).all()

        self.assertEqual(2, len(received_items))
        for item in received_items:
            self.assertLess(15, item["age"])

        #
        # Increment age
        #

        await self._client.kv.update(
            container=self._container, table_path=self._path, key="louise", expression="age = age + 1"
        )

        response = await self._client.kv.get(
            container=self._container, table_path=self._path, key="louise", attribute_names=["age"]
        )

        self.assertEqual(10, response.output.item["age"])

    async def test_limit(self):
        for idx in range(100):
            await self._client.kv.put(
                container=self._container,
                table_path=self._path,
                key=f"key-{idx}",
                attributes={
                    "attr": idx,
                },
            )

        # limit using all()
        received_items = await self._client.kv.new_cursor(
            container=self._container, table_path=self._path, limit=30
        ).all()

        self.assertEqual(len(received_items), 30)

    async def _delete_items(self, path, items):
        # delete items
        for item_key, _ in future.utils.viewitems(items):
            await self._client.kv.delete(container=self._container, table_path=path, key=item_key)

        # delete dir
        await self._client.object.delete(container=self._container, path=path)

    async def _verify_items(self, path, items):
        items_cursor = self._client.kv.new_cursor(container=self._container, table_path=path, attribute_names=["*"])

        received_items = await items_cursor.all()

        # TODO: verify contents
        self.assertEqual(len(items), len(received_items))

    def _compare_item_values(self, v1, v2):
        if isinstance(v1, array.array):
            # convert to list
            v1 = list(v1)

        if v1 != v2:
            self.fail("Values dont match")

    def _compare_item_types(self, v1, v2):
        if isinstance(v1, array.array):
            # convert to list
            v1 = list(v1)

        # can't guarantee strings as they might be converted to unicode
        if type(v1) is not str:
            self.assertEqual(type(v1), type(v2))


class TestRaiseForStatus(Test):
    def setUp(self):
        super(TestRaiseForStatus, self).setUp()

    async def test_always_raise_no_error(self):
        # should raise - since the status code is 500
        await self._client.container.list(
            self._container, "/", raise_for_status=v3io.dataplane.transport.RaiseForStatus.always
        )

    async def test_specific_status_code_match(self):
        # should raise - since the status code is 500
        await self._client.container.list(self._container, "/", raise_for_status=[200])

    async def test_specific_status_code_no_match(self):
        # should raise - since the status code is 500
        try:
            await self._client.container.list(self._container, "/", raise_for_status=[500])
        except Exception:
            return

        self.fail("Expected an exception")

    async def test_never_raise(self):
        await self._client.object.get(
            container=self._container, path="/non-existing", raise_for_status=v3io.aio.dataplane.RaiseForStatus.never
        )

    async def test_default_raise(self):
        try:
            await self._client.object.get(container=self._container, path="/non-existing")
        except Exception:
            return

        self.fail("Expected an exception")



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
class Cursor(object):
    def __init__(
        self,
        context,
        container_name,
        access_key,
        table_path,
        table_name,
        raise_for_status=None,
        attribute_names="*",
        filter_expression=None,
        marker=None,
        sharding_key=None,
        limit=None,
        segment=None,
        total_segments=None,
        sort_key_range_start=None,
        sort_key_range_end=None,
    ):
        self._context = context
        self._container_name = container_name
        self._access_key = access_key
        self._current_response = None
        self._current_items = None
        self._current_item = None
        self._current_item_index = 0
        self._total_items_read = 0

        # get items params
        self.raise_for_status = raise_for_status
        self.table_path = table_path
        self.table_name = table_name
        self.attribute_names = attribute_names
        self.filter_expression = filter_expression
        self.marker = marker
        self.sharding_key = sharding_key
        self.limit = limit
        self.segment = segment
        self.total_segments = total_segments
        self.sort_key_range_start = sort_key_range_start
        self.sort_key_range_end = sort_key_range_end

    async def next_item(self):
        calculated_limit = self.limit

        # check if we already reached the limit we were asked for
        if self.limit is not None:
            # if we already passed the limit, stop here
            if self._total_items_read >= self.limit:
                return None

            # don't ask for more items than we'll read
            calculated_limit -= self._total_items_read

        if self._current_item_index < len(self._current_items or []):
            self._current_item = self._current_items[self._current_item_index]
            self._current_item_index += 1
            self._total_items_read += 1

            return self._current_item

        if self._current_response and (self._current_response.output.last or len(self._current_items) == 0):
            return None

        self.marker = self._current_response.output.next_marker if self._current_response else None

        # get the next batch
        self._current_response = await self._context.kv.scan(
            self._container_name,
            self.table_path,
            self.table_name,
            self._access_key,
            self.raise_for_status,
            self.attribute_names,
            self.filter_expression,
            self.marker,
            self.sharding_key,
            calculated_limit,
            self.segment,
            self.total_segments,
            self.sort_key_range_start,
            self.sort_key_range_end,
        )

        # raise if there was an issue
        self._current_response.raise_for_status(self.raise_for_status)

        # set items
        self._current_items = self._current_response.output.items
        self._current_item_index = 0

        # and recurse into next now that we repopulated response
        return await self.next_item()

    async def all(self):
        items = []

        while True:
            item = await self.next_item()

            if item is None:
                break

            items.append(item)

        return items



# Copyright 2017 The Nuclio Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


class HumanReadableFormatter(logging.Formatter):
    def __init__(self):
        super(HumanReadableFormatter, self).__init__()

    def format(self, record):
        record_with = getattr(record, "with", {})
        if record_with:
            more = ": {0}".format(record_with)
        else:
            more = ""

        return "{0} [{1}] {2}{3}".format(
            self.formatTime(record, self.datefmt), record.levelname.lower(), record.getMessage(), more
        )


class Logger(object):
    def __init__(self, level="DEBUG"):
        self._logger = logging.getLogger("v3io_py")
        self._logger.setLevel(level)

    def set_handler(self, handler_name, file, formatter):
        # check if there's a handler by this name
        for handler in self._logger.handlers:
            if handler.name == handler_name:
                self._logger.removeHandler(handler)
                break

        # create a stream handler from the file
        stream_handler = logging.StreamHandler(file)
        stream_handler.name = handler_name

        # set the formatter
        stream_handler.setFormatter(formatter)

        # add the handler to the logger
        self._logger.addHandler(stream_handler)

    def debug(self, message, *args):
        self._logger.debug(message, *args)

    def info(self, message, *args):
        self._logger.info(message, *args)

    def warn(self, message, *args):
        self._logger.warning(message, *args)

    def error(self, message, *args):
        self._logger.error(message, *args)

    def debug_with(self, message, *args, **kw_args):
        self._logger.debug(message, *args, extra={"with": kw_args})

    def info_with(self, message, *args, **kw_args):
        self._logger.info(message, *args, extra={"with": kw_args})

    def warn_with(self, message, *args, **kw_args):
        self._logger.warning(message, *args, extra={"with": kw_args})

    def error_with(self, message, *args, **kw_args):
        self._logger.error(message, *args, extra={"with": kw_args})
